# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from google.events.cloud.dataflow import gapic_version as package_version

__version__ = package_version.__version__



from google.events.cloud.dataflow_v1beta3.types.data import AutoscalingSettings
from google.events.cloud.dataflow_v1beta3.types.data import BigQueryIODetails
from google.events.cloud.dataflow_v1beta3.types.data import BigTableIODetails
from google.events.cloud.dataflow_v1beta3.types.data import DatastoreIODetails
from google.events.cloud.dataflow_v1beta3.types.data import DebugOptions
from google.events.cloud.dataflow_v1beta3.types.data import Environment
from google.events.cloud.dataflow_v1beta3.types.data import ExecutionStageState
from google.events.cloud.dataflow_v1beta3.types.data import FileIODetails
from google.events.cloud.dataflow_v1beta3.types.data import Job
from google.events.cloud.dataflow_v1beta3.types.data import JobEventData
from google.events.cloud.dataflow_v1beta3.types.data import JobExecutionInfo
from google.events.cloud.dataflow_v1beta3.types.data import JobExecutionStageInfo
from google.events.cloud.dataflow_v1beta3.types.data import JobMetadata
from google.events.cloud.dataflow_v1beta3.types.data import Package
from google.events.cloud.dataflow_v1beta3.types.data import PubSubIODetails
from google.events.cloud.dataflow_v1beta3.types.data import SdkHarnessContainerImage
from google.events.cloud.dataflow_v1beta3.types.data import SdkVersion
from google.events.cloud.dataflow_v1beta3.types.data import SpannerIODetails
from google.events.cloud.dataflow_v1beta3.types.data import WorkerPool
from google.events.cloud.dataflow_v1beta3.types.data import AutoscalingAlgorithm
from google.events.cloud.dataflow_v1beta3.types.data import DefaultPackageSet
from google.events.cloud.dataflow_v1beta3.types.data import FlexResourceSchedulingGoal
from google.events.cloud.dataflow_v1beta3.types.data import JobState
from google.events.cloud.dataflow_v1beta3.types.data import JobType
from google.events.cloud.dataflow_v1beta3.types.data import ShuffleMode
from google.events.cloud.dataflow_v1beta3.types.data import TeardownPolicy
from google.events.cloud.dataflow_v1beta3.types.data import WorkerIPAddressConfiguration

__all__ = ('AutoscalingSettings',
    'BigQueryIODetails',
    'BigTableIODetails',
    'DatastoreIODetails',
    'DebugOptions',
    'Environment',
    'ExecutionStageState',
    'FileIODetails',
    'Job',
    'JobEventData',
    'JobExecutionInfo',
    'JobExecutionStageInfo',
    'JobMetadata',
    'Package',
    'PubSubIODetails',
    'SdkHarnessContainerImage',
    'SdkVersion',
    'SpannerIODetails',
    'WorkerPool',
    'AutoscalingAlgorithm',
    'DefaultPackageSet',
    'FlexResourceSchedulingGoal',
    'JobState',
    'JobType',
    'ShuffleMode',
    'TeardownPolicy',
    'WorkerIPAddressConfiguration',
)
