# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from google.events.cloud.datastream import gapic_version as package_version

__version__ = package_version.__version__



from google.events.cloud.datastream_v1.types.data import AvroFileFormat
from google.events.cloud.datastream_v1.types.data import BigQueryDestinationConfig
from google.events.cloud.datastream_v1.types.data import BigQueryProfile
from google.events.cloud.datastream_v1.types.data import ConnectionProfile
from google.events.cloud.datastream_v1.types.data import ConnectionProfileEventData
from google.events.cloud.datastream_v1.types.data import DestinationConfig
from google.events.cloud.datastream_v1.types.data import Error
from google.events.cloud.datastream_v1.types.data import ForwardSshTunnelConnectivity
from google.events.cloud.datastream_v1.types.data import GcsDestinationConfig
from google.events.cloud.datastream_v1.types.data import GcsProfile
from google.events.cloud.datastream_v1.types.data import JsonFileFormat
from google.events.cloud.datastream_v1.types.data import MysqlColumn
from google.events.cloud.datastream_v1.types.data import MysqlDatabase
from google.events.cloud.datastream_v1.types.data import MysqlProfile
from google.events.cloud.datastream_v1.types.data import MysqlRdbms
from google.events.cloud.datastream_v1.types.data import MysqlSourceConfig
from google.events.cloud.datastream_v1.types.data import MysqlSslConfig
from google.events.cloud.datastream_v1.types.data import MysqlTable
from google.events.cloud.datastream_v1.types.data import OracleColumn
from google.events.cloud.datastream_v1.types.data import OracleProfile
from google.events.cloud.datastream_v1.types.data import OracleRdbms
from google.events.cloud.datastream_v1.types.data import OracleSchema
from google.events.cloud.datastream_v1.types.data import OracleSourceConfig
from google.events.cloud.datastream_v1.types.data import OracleTable
from google.events.cloud.datastream_v1.types.data import PostgresqlColumn
from google.events.cloud.datastream_v1.types.data import PostgresqlProfile
from google.events.cloud.datastream_v1.types.data import PostgresqlRdbms
from google.events.cloud.datastream_v1.types.data import PostgresqlSchema
from google.events.cloud.datastream_v1.types.data import PostgresqlSourceConfig
from google.events.cloud.datastream_v1.types.data import PostgresqlTable
from google.events.cloud.datastream_v1.types.data import PrivateConnection
from google.events.cloud.datastream_v1.types.data import PrivateConnectionEventData
from google.events.cloud.datastream_v1.types.data import PrivateConnectivity
from google.events.cloud.datastream_v1.types.data import Route
from google.events.cloud.datastream_v1.types.data import RouteEventData
from google.events.cloud.datastream_v1.types.data import SourceConfig
from google.events.cloud.datastream_v1.types.data import StaticServiceIpConnectivity
from google.events.cloud.datastream_v1.types.data import Stream
from google.events.cloud.datastream_v1.types.data import StreamEventData
from google.events.cloud.datastream_v1.types.data import VpcPeeringConfig

__all__ = ('AvroFileFormat',
    'BigQueryDestinationConfig',
    'BigQueryProfile',
    'ConnectionProfile',
    'ConnectionProfileEventData',
    'DestinationConfig',
    'Error',
    'ForwardSshTunnelConnectivity',
    'GcsDestinationConfig',
    'GcsProfile',
    'JsonFileFormat',
    'MysqlColumn',
    'MysqlDatabase',
    'MysqlProfile',
    'MysqlRdbms',
    'MysqlSourceConfig',
    'MysqlSslConfig',
    'MysqlTable',
    'OracleColumn',
    'OracleProfile',
    'OracleRdbms',
    'OracleSchema',
    'OracleSourceConfig',
    'OracleTable',
    'PostgresqlColumn',
    'PostgresqlProfile',
    'PostgresqlRdbms',
    'PostgresqlSchema',
    'PostgresqlSourceConfig',
    'PostgresqlTable',
    'PrivateConnection',
    'PrivateConnectionEventData',
    'PrivateConnectivity',
    'Route',
    'RouteEventData',
    'SourceConfig',
    'StaticServiceIpConnectivity',
    'Stream',
    'StreamEventData',
    'VpcPeeringConfig',
)
