# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .data import (
    AdaptingOSStep,
    ApplianceVersion,
    AppliedLicense,
    AvailableUpdates,
    AwsSourceDetails,
    AwsSourceVmDetails,
    CloneJob,
    CloneJobEventData,
    CloneStep,
    ComputeEngineTargetDefaults,
    ComputeEngineTargetDetails,
    ComputeScheduling,
    CutoverForecast,
    CutoverJob,
    CutoverJobEventData,
    CutoverStep,
    CycleStep,
    DatacenterConnector,
    DatacenterConnectorEventData,
    Group,
    GroupEventData,
    InitializingReplicationStep,
    InstantiatingMigratedVMStep,
    MigratingVm,
    MigratingVmEventData,
    MigrationWarning,
    NetworkInterface,
    PostProcessingStep,
    PreparingVMDisksStep,
    ReplicatingStep,
    ReplicationCycle,
    ReplicationSync,
    SchedulePolicy,
    SchedulingNodeAffinity,
    ShuttingDownSourceVMStep,
    Source,
    SourceEventData,
    TargetProject,
    TargetProjectEventData,
    UpgradeStatus,
    UtilizationReport,
    UtilizationReportEventData,
    VmUtilizationInfo,
    VmUtilizationMetrics,
    VmwareSourceDetails,
    VmwareVmDetails,
    ComputeEngineBootOption,
    ComputeEngineDiskType,
    ComputeEngineLicenseType,
)

__all__ = (
    'AdaptingOSStep',
    'ApplianceVersion',
    'AppliedLicense',
    'AvailableUpdates',
    'AwsSourceDetails',
    'AwsSourceVmDetails',
    'CloneJob',
    'CloneJobEventData',
    'CloneStep',
    'ComputeEngineTargetDefaults',
    'ComputeEngineTargetDetails',
    'ComputeScheduling',
    'CutoverForecast',
    'CutoverJob',
    'CutoverJobEventData',
    'CutoverStep',
    'CycleStep',
    'DatacenterConnector',
    'DatacenterConnectorEventData',
    'Group',
    'GroupEventData',
    'InitializingReplicationStep',
    'InstantiatingMigratedVMStep',
    'MigratingVm',
    'MigratingVmEventData',
    'MigrationWarning',
    'NetworkInterface',
    'PostProcessingStep',
    'PreparingVMDisksStep',
    'ReplicatingStep',
    'ReplicationCycle',
    'ReplicationSync',
    'SchedulePolicy',
    'SchedulingNodeAffinity',
    'ShuttingDownSourceVMStep',
    'Source',
    'SourceEventData',
    'TargetProject',
    'TargetProjectEventData',
    'UpgradeStatus',
    'UtilizationReport',
    'UtilizationReportEventData',
    'VmUtilizationInfo',
    'VmUtilizationMetrics',
    'VmwareSourceDetails',
    'VmwareVmDetails',
    'ComputeEngineBootOption',
    'ComputeEngineDiskType',
    'ComputeEngineLicenseType',
)
