# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

from deprecated import deprecated

STATE: Final = "state"
"""
Deprecated: Replaced by `db.client.connection.state`.
"""


@deprecated(reason="The attribute state is deprecated - Replaced by `db.client.connection.state`")  # type: ignore
class StateValues(Enum):
    IDLE = "idle"
    """idle."""
    USED = "used"
    """used."""
