# -*- coding: utf-8 -*-
"""
Deprecated Library
==================

Python ``@deprecated`` decorator to deprecate old python classes, functions or methods.

"""

__version__ = "1.2.18"
__author__ = u"Laurent LAPORTE <laurent.laporte.pro@gmail.com>"
__date__ = "2025-01-27"
__credits__ = "(c) Laurent LAPORTE"

from deprecated.classic import deprecated
