# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from google.events.cloud.apigeeregistry_v1 import gapic_version as package_version

__version__ = package_version.__version__



from .types.data import Api
from .types.data import ApiDeployment
from .types.data import ApiDeploymentEventData
from .types.data import ApiEventData
from .types.data import ApiSpec
from .types.data import ApiSpecEventData
from .types.data import ApiVersion
from .types.data import ApiVersionEventData
from .types.data import Instance
from .types.data import InstanceEventData

__all__ = (
'Api',
'ApiDeployment',
'ApiDeploymentEventData',
'ApiEventData',
'ApiSpec',
'ApiSpecEventData',
'ApiVersion',
'ApiVersionEventData',
'Instance',
'InstanceEventData',
)
